/* fepskkfep.c (based on feponew.c) */					/* vi:set ts=4 sw=4: */
/*
 * Author:
 *	  KIHARA Hideto <deton@m1.interq.or.jp>
 *	  $Id: fepskkfep.c,v 1.15 2000/12/21 12:18:51 deton Exp $
 */

#ifdef SKKFEP
#include "vim.h"
#include "globals.h"
#include "proto.h"
#include "param.h"
#include "kanji.h"

#ifdef FEP_TRACE
# define		DTRC(_1, _2, _3, _4)	skkfepTrace(_1, _2, _3, _4)
#else
# define		DTRC(_1, _2, _3, _4)
#endif
/*#define NORUBOUT  /* don't clear chars temporary written. only move cursor */

#ifdef FEP_TRACE
static void		skkfepTrace __PARMS((char_u *, char_u *, int, int));
#endif
static void		skkfep_msg __PARMS((char_u *, int *, int *));
static void		skkfep_clrmsg __PARMS((void));
static void		restore_vimflags  __PARMS((void));

/* variables visible to external */
static int				skkfep_mode = FALSE;		/* actual conversion mode */

/* input controls */
#define MAXBUF			1024
static unsigned char	outbuf[MAXBUF];
static int				ohead = 0;
static int				otail = 0;
static unsigned char	wrkbuf[MAXBUF];
static unsigned char	delay_buf[MAXBUF];

/* internal controls */
static int				ready = FALSE;	/* initialized? */

/* screen controls */
static int				visualmode = FALSE;
static int				disp_row;
static int				disp_col;
static int				mline  = FALSE;
static int				sline  = FALSE;
static int				cline  = FALSE;
	/* if line was continued with '\\' {{ only support one '\\' }} */
static int				line_continued = 0;

/* to link with skkfep */
#include "config.h"	/* skkfep's config.h */
#ifndef MOVE_CURSOR
# error "Please #define MOVE_CURSOR in skkfep's config.h to link with jvim3"
#endif
#ifdef showmode		/* function name conflicts between skkfep and jvim */
# undef showmode	/* so define showmode as showSKKmode in skkfep/config.h */
#endif				/* Because this fepskkfep.c is source of jvim, undef it */
#include "fep.h"
#include "functions.h"
KeymapPtr CurrentKeymap;
char UserDicName[64];
Dictionary UserDic = NULL;
#ifdef SYSTEM_DIC_NAME
Dictionary SystemDic;
#endif
#ifdef USE_SERVER
int ServerOK = 0;
#endif
#ifdef BOTH_SERVER_AND_SKKSRCH
int (*openSKKservPtr)() = openSKKserv;
CandList (*getCandFromServerPtr)() = getCandFromServer;
int (*closeSKKservPtr)() = closeSKKserv;
#endif
short BlockTty = 0;
char PreserveOnFailure = PRESERVE_ON_FAILURE;
char OkuriFirst = 0;
static int skkfep_lastmode = SKK_MODE;
static int skkfep_curmode = KANA_MODE;
static int skkfep_lastshowmode = SKK_MODE;
static int skkfep_flush = FALSE;
static int skkfep_injecting = FALSE;	/* same as getchar.c:FepIn */

/* display SKK mode using jvim's showmode() */
	void
showSKKmode(s)
	int s;
{
    skkfep_lastmode = skkfep_curmode;
    skkfep_curmode = s;
	switch (s) {
		case SKK_MODE:
		case KANA_MODE:
		case KKANA_MODE:
		case ZENEI_MODE:
		case CODEIN_MODE:
			if (s != skkfep_lastshowmode) {
				skkfep_lastshowmode = s;
				showmode();		/* vim's function */
				DTRC("@showSKKmode", "(%d, %d)", disp_row, disp_col);
				if (!RedrawingDisabled)
					windgoto(disp_row, disp_col);
			}
			break;
		default:	/* do not show other mode */
			break;
	}
}

	void
showlastmode()
{
    showSKKmode(skkfep_lastmode);
    skkfep_curmode = skkfep_lastmode;
}

/* temporary display some characters like candidates */
	int
writes(s)
	char *s;
{
	/* must flush out outbuf before print s */
	/* ex:if typed "tabunDaijoubu", must not print "\/d" before flushOut "nn" */
	if (skkfep_flush && skkfep_isready()) {
		strncat(delay_buf, s, sizeof(delay_buf) - strlen(delay_buf) - 1);
	} else {
		skkfep_flush = FALSE;
		if (delay_buf[0] != NUL) {
			skkfep_msg(delay_buf, &disp_row, &disp_col);
			delay_buf[0] = NUL;
		}
		skkfep_msg(s, &disp_row, &disp_col);
	}
}

	int
write1(c)
	char c;
{
	char buf[2];
	buf[0] = c;
	buf[1] = NUL;
	writes(buf);
}

/* put character on outbuf to return by skkfep_getkey() */
	int
writeShell1(c)
	char c;
{
#ifdef REGMODE
	if (Registering)
		regUpdateCand1(c);
	else
#endif /* REGMODE */
	if (otail >= MAXBUF - 1)
	{
		emsg("Input buffer filled.");
		return 0;
	}
	else
	{
		outbuf[otail++] = c;
		outbuf[otail] = NUL;
	}
	return 1;
}

/* put string on outbuf to return by skkfep_getkey() */
	int
writeShells(s)
	char *s;
{
#ifdef REGMODE
	if (Registering)
		regUpdateCands(s);
	else
#endif /* REGMODE */
	{
		DTRC("@writeShells:O", "[%s]", (int)s, 0);
		otail += kanjiconvsfrom(s, STRLEN(s),
						&outbuf[otail], sizeof(outbuf) - otail,
						NULL, JP_EUC, NULL);
		outbuf[otail] = NUL;
	}
	return strlen(s);
}

	int
msgLine()
{
	return 0;	/* {{ not yet support to use message line }} */
}

	void
toMsg()
{
}

	void
fromMsg()
{
}

	void
csrLeft(n)
	int n;
{
	DTRC("@csrLeft:", "[%d] %d", n, disp_col);
	if (n <= 0)
		return;
	if (disp_col >= n)
	{
		disp_col -= n;
		if (cursor_left(n) < 0)	/* if failed (not have "le") */
		{
			windgoto(disp_row, disp_col);	/* use windgoto */
			DTRC("@>csrLeft:", "windgoto(row=%d, col=%d)", disp_row, disp_col);
		}
		return;
	}
	/* must go to upper line */
	while (n > 0)
	{
		disp_col -= n;
		if (disp_col >= 0)	/* done */
			break;
		/* here disp_col < 0 */
		n = -disp_col;	/* must go left for -disp_col */
		disp_col = Columns;
		if (line_continued)
			--disp_col;
		if (--disp_row < 0)	/* cannot go to upper more one line */
		{
			disp_row = 0;
			disp_col = 0;
			break;
		}
	}
	windgoto(disp_row, disp_col);
	DTRC("@>csrLeft:", "(row=%d, col=%d)", disp_row, disp_col);
}

	void
csrRight(n)
	int n;
{
	DTRC("@csrRight:", "[%d] %d", n, disp_col);
	if (n <= 0)
		return;
	if (disp_col + n < Columns
	&& !(line_continued && disp_col + n == Columns - 1))	/* at '\\' */
	{
		disp_col += n;
		if (cursor_right(n) < 0)	/* if failed (not have "nd") */
		{
			windgoto(disp_row, disp_col);	/* use windgoto */
			DTRC("@>csrRight:", "windgoto(row=%d, col=%d)", disp_row, disp_col);
		}
		return;
	}
	/* must go to lower line */
	while (n > 0)
	{
		disp_col += n;
		if (disp_col >= Columns - 1 && line_continued)
			++disp_col;
		if (disp_col < Columns)	/* done */
			break;
		/* here disp_col >= Columns */
		n = disp_col - Columns;	/* must go right for rest n */
		disp_col = 0;
		if (++disp_row >= Rows)	/* cannot go to lower more one line */
		{
			disp_row = Rows - 1;
			disp_col = Columns - 1;
			break;
		}
	}
	windgoto(disp_row, disp_col);
	DTRC("@>csrRight:", "(row=%d, col=%d)", disp_row, disp_col);
}

	void
erase(n)
	int n;
{
	int lc = line_continued;
	int toleft = 0;
	DTRC("@erase:", "[%d] %d", n, disp_col);
	if (n <= 0)
		return;
	while (n) {
		write1(' ');
		/* if overwrite '\\' which marks line continued,
		 * write one more ' ' to erase real character. {{ only support one \ }}
		 */
		if (lc && !line_continued)
		{
			write1(' ');
			lc = line_continued;
			toleft = 1;
		}
		n--;
	}
	if (toleft)	/* fixes position */
		csrLeft(1);
	DTRC("@>erase:", "(%d, %d)", disp_col, disp_row);
}

	void
rubout(n)
	int n;
{
	DTRC("@rubout:", "[%d] %d", n, disp_col);
	if (n <= 0)
		return;
	csrLeft(n);
#ifndef NORUBOUT
	erase(n);
	csrLeft(n);
#endif
	DTRC("@>rubout:", "(%d, %d)", disp_col, disp_row);
}

	void
bell()
{
	beep();
}

	void
flushOut(n)
	int n;
{
#ifdef REGMODE
	if (!Registering)	/* if Registering, output is to writes() */
#endif
		skkfep_flush = TRUE;	/* flush out outbuf before print in writes() */
}

void
nulcmd(c)
char c;
{}

void
thru(c)
char c;
{
	writeShell1(c);
}

void
toAsc(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,convertKeymap(&NormalKeymap));
	showSKKmode(SKK_MODE);
#ifdef REGMODE
	if (!Registering)		/* if Registering, must check ASCII, too */
#endif
		skkfep_mode = FALSE;	/* ASCII mode and quit skkfep_inject() */
}

void
toZenA(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,ZenkakuKeymap);
	showSKKmode(ZENEI_MODE);
}

void
thruToAsc(c)
char c;
{
	toAsc(c);
	thru(c);
}

void
toEsc(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,convertKeymap(&EscapedKeymap));
	showSKKmode(SKK_MODE);
}

void
thruToEsc(c)
char c;
{
	toEsc(c);
	thru(c);
}

void
thruBack(c)
char c;
{
	thru(c);
	toKana(c);
}

void
thru1(c)
char c;
{
	thru(c);
	restoreKeymap(&CurrentKeymap);
	showlastmode();
}

/* set SKKFEP keymap */
	static void
skkfep_setkeymap(skkmode)
	int skkmode;
{
	DTRC("skkfep_setkeymap: ", "%d", skkmode, 0);
	switch (skkmode) {
		case SKK_MODE:		/* main loop may have been ended by 'l' */
			toKana();
			break;
		case KANA_MODE:
		case KKANA_MODE:
			toKana();		/* Hira/Kata */
			break;
		case ZENEI_MODE:	/* toZenA() */
			setKeymap(&CurrentKeymap,ZenkakuKeymap);
			showSKKmode(skkmode);
			break;
		/* following cases may not occur */
		case KINPUT_MODE:	/* kkBeg() but do not WordBufLen = 0 */
			setKeymap(&CurrentKeymap,KanjiInputKeymap);
			break;
		case KALPHA_INPUT_MODE:	/* kkBegA() but do not WordBufLen = 0 */
			setKeymap(&CurrentKeymap,KAlphaInputKeymap);
			break;
		default:
			toKana();
			break;
	}
}

#ifdef FEP_TRACE
/*------------------------------------------------------------*/
static void
skkfepTrace(title, fmt, a1, a2)
char_u	*	title;
char_u	*	fmt;
int		 	a1, a2;
{
	static FILE		*fp = NULL;

	if (fp == NULL)
	{
		if ((fp = fopen("skkfep.trc", "w+")) == NULL)
			return;
		fputs("start trace\n", fp);
	}
	fprintf(fp, "%s ", title);
	fprintf(fp, fmt, a1, a2);
	fprintf(fp, "\n");
}
#endif /* FEP_TRACE */

static void
skkfep_msg(str, row, col)
char_u	*	str;
int		*	row;
int		*	col;
{
	int	len = 0;
	char_u *msg;
	int setpos = 0;

	DTRC("@skkfep_msg", "%s:%x", (int)str, *str);
	len = kanjiconvsfrom(str, STRLEN(str), wrkbuf, sizeof(wrkbuf),
															NULL, JP_EUC, NULL);
	wrkbuf[len] = NUL;
	if (wrkbuf[0] == NUL)
	{
		return;
	}

	msg = wrkbuf;
	while (*msg)
	{
		if (line_continued && *col == 0 && *row > 0 && !ISkanji(*msg))
		{	/* if writing at head of line and previous line
			 * is terminated with "\\" which marks line continue,
			 * then write from tail of previous line and overwrite '\\'.
			 */
			--*row;
			*col = Columns - 1;
			line_continued = 0;
			setpos = 1;
		}
		screen_msg(msg, *row, *col);
		while (*msg && *col < Columns)
		{
			if (ISkanji(*msg))
			{
				if (*col == Columns - 1)
				{
					screen_msg("\\", *row, *col);
					*row += 1;
					*col  = 0;
					line_continued = 1;
					break;
				}
				else if (line_continued && *col == Columns - 2)
				{	/* "\\" which marks line continue is overwritten 
					 * by some char
					 */
					line_continued = 0;
				}
				*col += 1;
				++msg;
			}
			/* not Kanji */
			else if (line_continued && *col == Columns - 1) 
			{	/* "\\" which marks line continue is overwritten by some char */
				line_continued = 0;
			}
			*col += 1;
			++msg;
		}
		if (*col >= Columns)
		{
			*col  = 0;
			*row += 1;
		}
		if (*row >= Rows)
			break;
	}
	if (setpos)
		windgoto(*row, *col);
	flushbuf();
	DTRC("@>skkfep_msg:", "(%d, %d)", *col, *row);
}

static void
skkfep_clrmsg()
{
	DTRC("@skkfep_clrmsg", "", 0, 0);
	if (visualmode)
	{
		gotocmdline(TRUE, NUL);
		windgoto(Rows - 1, 0);
		if (T_EL != NULL && *T_EL != NUL)
			outstr(T_EL);
		else
		{
			memset(wrkbuf, ' ', Columns);
			wrkbuf[Columns - 1] = NUL;
			outstr(wrkbuf);
		}
	}
	else
	{
		WIN		*	wp;
		WIN		*	twp;
		linenr_t	lnum;
		linenr_t	tlnum;

		twp = firstwin;
		while (twp != NULL)
		{
			if (twp->w_winpos == 0)
				break;
			twp = twp->w_next;
		}
		if (twp == NULL)
			return;
		tlnum = twp->w_topline;
		wp = curwin;
		lnum = curwin->w_cursor.lnum;

		win_enter(twp, TRUE);
		curwin->w_cursor.lnum = tlnum;
		windgoto(0, 0);
		if (T_EL != NULL && *T_EL != NUL)
			outstr(T_EL);
		RedrawingDisabled++;
		cursupdate();
		RedrawingDisabled--;
		setcursor();
		screen_fill(0, 1, 0, (int)Columns, ' ', ' ');
		updateline();

		win_enter(wp, TRUE);
		curwin->w_cursor.lnum = lnum;
	}
	flushbuf();
}

#if defined(USE_SERVER) && defined(SIGPIPE)
static SignalFunc
sigpipe_handler()
{
	SignalFunc (*sigfunc)();

	ServerOK = 0;
	sigfunc = signal(SIGPIPE, SIG_IGN);
# ifdef BOTH_SERVER_AND_SKKSRCH
	closeSKKservPtr();
# else
	closeSKKserv();
# endif
	signal(SIGPIPE, sigfunc);
	return;
}
#endif	/* USE_SERVER */

/*------------------------------------------------------------*/
/*
 * internal interface for skkfep
 */

/* inject key stroke into skkfep, and fill reply buffer if needed. */
void
skkfep_inject(ch)
int			ch;
{
	int	len, i, row, col;
	int	c = NUL, o;

	DTRC("@skkfep_inject:S", "", 0, 0);
	skkfep_injecting = TRUE;
	if (visualmode)
	{
		RedrawingDisabled++;
		cursupdate();
		RedrawingDisabled--;
		setcursor();
		updateline();
		row = curwin->w_winpos + curwin->w_row;
		col = curwin->w_col;
	}
	else
	{
		redrawcmdline();
		row = msg_row;
		col = msg_col;
	}
	DTRC("@skkfep_inject:P", "%d:%d", row, col);
	if (skkfep_flush && delay_buf[0] != NUL)
	{
		DTRC("@skkfep_inject:B", "%s", (int)delay_buf, 0);
		len = strlen(delay_buf);
		skkfep_msg(delay_buf, &row, &col);
		delay_buf[0] = NUL;
		skkfep_flush = FALSE;
		DTRC("@skkfep_inject:X", "%d:%d", len, disp_col);
		if (visualmode)
		{
			cline = sline = mline = FALSE;
			if (row >= (curwin->w_winpos + curwin->w_height))
			{
				if (row >= Rows - 1)
					cline = TRUE;
				else
					sline = TRUE;
			}
			for (i = 0, len = 0; i < Rows; i++)
			{
				if (curwin->w_lsize_lnum[i] == curwin->w_cursor.lnum)
				{
					DTRC("@BS:len", "%d:%d", len, curwin->w_winpos);
					DTRC("@BS:row", "%d:%d", row, curwin->w_lsize[i]);
					len += curwin->w_winpos;
					if (curwin->w_lsize[i] == 0 && len == row)
						mline = FALSE;
					else if (curwin->w_lsize[i] == 0)
						mline = TRUE;
					else if (len <= row && row <= (len + curwin->w_lsize[i] -1))
						mline = FALSE;
					else
						mline = TRUE;
					DTRC("@BS:cline", "%d:%d", cline, sline);
					DTRC("@BS:mline", "%d:%d", mline, curwin->w_height);
					break;
				}
				else
					len += curwin->w_lsize[i];
			}
		}
		disp_row = row;
		disp_col = col;
	}
	else if (visualmode)
	{
		disp_row = curwin->w_winpos + curwin->w_row;
		disp_col = curwin->w_col;
	}
	else
	{
		disp_row = msg_row;
		disp_col = msg_col;
	}
	line_continued = 0;

	/* SKKFEP main loop (get key input) */
	/* return injected Japanese characters as soon as possible
	 * ex: if typed "a" in Kana mode,
	 *     inject "a"(in kana) in outbuf and return immediately.
	 */
	while (skkfep_mode && !skkfep_isready()) {
		o = OkuriFirst;
		OkuriFirst = 0;
		if (ch != 0) {
			c = ch;
			ch = 0;
		}
		else {
			c = vgetc();
		}
		DTRC("inject: ", "%x", c, 0);
		if (ISkanji(c)) { /* this occurs by redo command. (KeyTyped = FALSE?) */
			thru(c);		/* inject into outbuf */
			c = vgetc();	/* must get 2-byte kanji char as a pair */
			DTRC("inject2: ", "%x", c, 0);
			thru(c);		/* inject into outbuf */
			break;
			/* break while loop because now skkfep_isready() */
		}
		else if (ISkana(c) || c < 0 || c > 0x7f) {	/* HANKAKU-KANA | illegal */
			thru(c);
			break;
		}
		else	/* must be 0 <= c <= 0x7f for Keymap */
			(*CurrentKeymap[c])(c,o);
	}
	/* there are two ways to quit above main loop:
	 * 1) !skkfep_mode: typed 'l' or ESC when SimpleEsc mode.
	 * 2) skkfep_isready(): injected some Japanese characters in outbuf.
	 */
	if (c == ESC) {	/* if typed ESC when skkfep_isready(),skkfep_mode != FALSE*/
		DTRC("inject: c==ESC: ", "cur=%d, last=%d", skkfep_curmode, skkfep_lastmode);
		skkfep_curmode = skkfep_lastmode;	/* if SimpleEsc, curmode is Asc */
		skkfep_mode = TRUE;
#ifdef NORUBOUT
		updateline();
#endif
	}
	else if (skkfep_mode == FALSE) {
		restore_vimflags();
	}
	skkfep_injecting = FALSE;
}

/* check if skkfep reply buffer is filled. */
int
skkfep_isready()
{
	return(ohead < otail);
}

/* fetch a character from skkfep reply buffer. */
int
skkfep_getkey()
{
	int			ret;

	if (!ready)
		return(-1);
	if (ohead >= otail)
		return(-1);		/* outbuf is empty */
	ret = outbuf[ohead++];
	DTRC("@skkfep_getkey", "%x", ret, 0);
	if (ohead >= otail)
	{
		ohead = 0;
		otail = 0;
		outbuf[0] = NUL;
		DTRC("@skkfep_getkey:clear", "%x", ret, 0);
		if (cline)
			skkfep_clrmsg();
		if (sline)
			win_redr_status(curwin);
	}
	return(ret);
}

/* Switch the screen control mode.
 * flag = FALSE: switch to dumb mode. (for ex mode)
 * flag = TRUE:  switch to visual mode. (for vi input mode)
 */
void
skkfep_visualmode(flag)
int			flag;
{
	if (!ready)
		return;
	visualmode = flag;
}

/* resize notification */
void
skkfep_resize()
{
	skkfep_visualmode(visualmode);
}

/*------------------------------------------------------------*/
/*
 * external fepctrl-based interface for skkfep
 */

static	int	keepmode = FALSE;	/* skkfep is active in input mode?
								 * (skkfep will be active when fep_on())
								 */

int
skkfep_init()
{
	char *env;
	if (ready)
		return(ready);
	ready = TRUE;

	setEscape(SimpleEsc);
	LastEscapeBehavior = NoEsc;
	setKanaKey();
	/* to be able to change LearnMode on MS-DOS */
	if ((env = getenv("SKKFEP_LEARNMODE")) != NULL) {
		switch (*env) {
			case 'o': /* off */
			case 'n': /* none */
			case '0':
				NetLearnMode = LearnOff;
				break;
			case 't': /* temporal */
			case '1':
				NetLearnMode = LearnTemporal;
				break;
			case 'p':
			case '2':
				NetLearnMode = LearnPermanent;
				break;
			default:
				NetLearnMode = NET_LEARN_MODE;
				break;
		}
	}
	if (*UserDicName == '\0') {
		if ((env = getenv("HOME")) != NULL) {
			strcpy(UserDicName,env);
			strcat(UserDicName,PATHSEPSTR);
		}
		strcat(UserDicName,USER_DIC_NAME);
	}
	if (UserDic == NULL) {
		if ((UserDic = openSKK(UserDicName)) == NULL)
			emsg("openSKK(UserDicName) failed.");
	}
#ifdef USE_SERVER
	if (ServerOK == 0) {
		if (openSKKserv() != -1)
		{
# ifdef BOTH_SERVER_AND_SKKSRCH
			openSKKservPtr = openSKKserv;
			getCandFromServerPtr = getCandFromServer;
			closeSKKservPtr = closeSKKserv;
# endif
			ServerOK = 1;
# ifdef SIGPIPE
			signal(SIGPIPE, sigpipe_handler);
# endif
		}
# ifdef BOTH_SERVER_AND_SKKSRCH
		else if (openSKKsrch() != -1) {
			openSKKservPtr = openSKKsrch;
			getCandFromServerPtr = getCandFromSKKsrch;
			closeSKKservPtr = closeSKKsrch;
			ServerOK = 2;	/* use skksrch.c */
		}
# endif
	}
	if (ServerOK == 0)
		emsg("Failed to access global SKK JISYO");
#endif
	DTRC("@skkfep_init", "%x", ready, 0);
	return(ready);
}

void
skkfep_term()
{
	if (!ready)
		return;
	DTRC("@skkfep_term", "", 0, 0);
#ifdef USE_SERVER
	if (ServerOK) {
# ifdef BOTH_SERVER_AND_SKKSRCH
		closeSKKservPtr();
# else
		closeSKKserv();
# endif
		ServerOK = 0;
		ready = FALSE;	/* to re-openSKKserv() in skkfep_init() */
	}
# ifdef SIGPIPE
	signal(SIGPIPE, SIG_DFL);	/* ServerOK may become 0 in sigpipe_handler() */
# endif
#endif

#ifdef USERDIC_CHANGED_CHECK
	if (UserDic == NULL || !UserDic_Changed)
#else
	if (UserDic == NULL)
#endif
		return;

#ifdef USE_SERVER
	/* if any candidates were registered, 
	 * write UserDic regardless of NetLearnMode (even if LearnOff).
	 */
# if 0 /*ifdef USERDIC_CHANGED_CHECK {{if LearnOff, should not save for FD?}}*/
	if (NetLearnMode == LearnPermanent || (UserDic_Changed & UDIC_REGISTER))
# else
	if (NetLearnMode == LearnPermanent)
# endif
#endif
	{
		msg("Saving SKK JISYO...");
		closeSKK(UserDic,UserDicName);	/* save and free() UserDic */
		UserDic = NULL;
#ifdef USERDIC_CHANGED_CHECK
		UserDic_Changed = 0;
#endif
		msg("");
		ready = FALSE;
	}
	/* if do not closeSKK(), UserDic is ready because it is not free()'ed */
}

void
skkfep_on()
{
	if (!ready)
		return;
	DTRC("@skkfep_on", "%x", keepmode, 0);
	if (keepmode)
		skkfep_force_on();
}

void
skkfep_off()
{
	if (!ready)
		return;
	DTRC("@skkfep_off", "%x", keepmode, 0);
	keepmode = skkfep_mode;
	skkfep_force_off();
}

void
skkfep_force_on()
{
	if (!ready || skkfep_mode)
		return;
	DTRC("@skkfep_force_on", "", 0, 0);
	skkfep_mode = TRUE;
	skkfep_setkeymap(skkfep_curmode);	/* set keymap and show SKK mode */
}

static void
restore_vimflags(void)
{
	if (!visualmode)
	{
		skkfep_clrmsg();
		redrawcmdline();
	}
}

void
skkfep_force_off()
{
	int saved_skkfep_curmode = skkfep_curmode;
	if (!ready || !skkfep_mode)
		return;
	DTRC("@skkfep_force_off", "", 0, 0);
	if (skkfep_curmode != SKK_MODE)
	{
		toAsc();
		skkfep_mode = FALSE;
	}
	restore_vimflags();
	skkfep_curmode = saved_skkfep_curmode;
}

int
skkfep_get_mode()
{
	if (!ready)
		return(0);
	DTRC("@skkfep_get_mode", "%x", skkfep_mode, 0);
	return(skkfep_mode);
}

/*
 *	skkfep_modestring() return FEP mode string
 */
char *
skkfep_get_modestring()
{
	if (!ready)
		return "";
	/* use skkfep_lastshowmode instead of skkfep_curmode.
	 * This function is used only in showmode() to show SKK mode,
	 * so it is ok to return skkfep_lastshowmode only. (see showSKKmode())
	 */
	DTRC("@skkfep_get_modestring", "%i:%s", 
			skkfep_lastshowmode, (int)modeString[skkfep_lastshowmode]);
	return modeString[skkfep_lastshowmode];
}

/*
 * return TRUE if in fep_inject(), FALSE for otherwise.
 */
int
skkfep_isinjecting()
{
	return skkfep_injecting;
}

/*
 * skkfep_freqsave();	save dictionary contents
 */
void
skkfep_freqsave()
{
}

#endif	/*SKKFEP*/
