/*
 *	SKK server support
 *
 *	by A.ITO	3 Jul, 1992
 */
#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include "skklib.h"
#include "fep.h"

#ifdef USE_SERVER

#ifdef WIN32
#include <winsock.h>
#include <io.h>	/* for prototypes of read(), write(), etc. */
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define closesocket(s)	close(s)
#endif

#define SKK_PORT_NUMBER	1178
#define SKK_SERVICENAME	"skkserv"

#ifdef BCOPY
#define bincopy(a,b,n)	bcopy(a,b,n)
#define fillzero(a,n)	bzero(a,n)
#else
#define bincopy(a,b,n)	memcpy(b,a,n)
#define fillzero(a,n)	memset(a,0,n)
#endif

#ifdef WIN32
static int win32skkservsock = -1;
#endif
static int skkservsock = -1;
static FILE *rserv,*wserv;
char *SKKServerHost = NULL;

char *getenv();

openSKKserv()
{
	int sock;
	int i;
	unsigned short port;
	struct sockaddr_in hostaddr;
	struct hostent *entry;
	struct servent *serv;
	struct protoent *proto;
	int a1,a2,a3,a4;
	char *hostname;

	if (SKKServerHost)
		hostname = SKKServerHost;
	else if ((hostname = getenv("SKKSERVER")) == NULL) {
#ifdef SKK_SERVER_HOST
		hostname = SKK_SERVER_HOST;
#else
		return -1;
#endif
	}
	if (strcmp(hostname, "!skkserv") == 0) {
		return -1;	/* specified not to use skkserv */
	}

#ifdef WIN32
	{
		WORD wVersionRequested;
		WSADATA wsaData;
		int sockopt = SO_SYNCHRONOUS_NONALERT;
		wVersionRequested = MAKEWORD(1, 1);
		if (WSAStartup(wVersionRequested, &wsaData) != 0) {
			return -1;
		}
		/* Enable the use of sockets as filehandles */
		setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE,
			   (char *)&sockopt, sizeof(sockopt));
	}
#endif
	serv = getservbyname(SKK_SERVICENAME,"tcp");
	fillzero((char*)&hostaddr,sizeof(struct sockaddr_in));
	if ((proto = getprotobyname("tcp")) == NULL) {
		return -1;
	}

	if ('0' <= *hostname && *hostname <= '9') {
		if (sscanf(hostname,"%d.%d.%d.%d",&a1,&a2,&a3,&a4) != 4) {
			return -1;
		}
		a1 = (a1<<24)|(a2<<16)|(a3<<8)|a4;
		hostaddr.sin_addr.s_addr = htonl(a1);
	}
	else {
		if ((entry = gethostbyname(hostname)) == NULL) {
			return -1;
		}
		bincopy(entry->h_addr, &hostaddr.sin_addr, entry->h_length);
	}

	if ((sock = socket(AF_INET,SOCK_STREAM,proto->p_proto)) < 0) {
		return -1;
	}
        hostaddr.sin_family = AF_INET;
	hostaddr.sin_port = serv ? serv->s_port : htons(SKK_PORT_NUMBER);
	if (connect(sock,(struct sockaddr *)&hostaddr,sizeof(struct sockaddr_in)) < 0) {
		closesocket(sock);
		return -1;
	}
#ifndef SKKFEP_LIB
	printf("SKKSERVER=%s\r\n",hostname);
#endif
#ifdef WIN32
	if ((skkservsock = _open_osfhandle(sock, 0)) < 0) {
		closesocket(sock);
    		return -1;
	}
	win32skkservsock = sock;
	rserv = fdopen(skkservsock,"rb");
	wserv = fdopen(skkservsock,"wb");
#else
	skkservsock = sock;
	rserv = fdopen(sock,"r");
	wserv = fdopen(sock,"w");
#endif
	return 0;
}

CandList
getCandFromServer(s)
char *s;
{
	char r;
	CandList res;

	fprintf(wserv,"1%s \n",s);
	fflush(wserv);
	if (read(skkservsock,&r,1) <= 0) {
		extern int ServerOK;
		SignalFunc (*sigfunc)();

		ServerOK = 0;
#ifdef SIGPIPE	/* WIN32 may not define SIGPIPE */
		sigfunc = signal(SIGPIPE, SIG_IGN);
#endif
		closeSKKserv();
#ifdef SIGPIPE
		signal(SIGPIPE, sigfunc);
#endif
		return NULL;
	}
	if (r == '1') {  /* succeeded */
		res = getCandList(rserv,NULL,isConjugate(s,strlen(s)));
		return res;
	}
	else
		while (read(skkservsock,&r,1) > 0 && r != '\n');
	return NULL;
}

closeSKKserv()
{
	if (skkservsock >= 0) {
		fprintf(wserv,"0\n");
		fflush(wserv);
		fclose(wserv);
		fclose(rserv);
		close(skkservsock);
		skkservsock = -1;
#ifdef WIN32
		closesocket(win32skkservsock);
		win32skkservsock = -1;
		WSACleanup();
#endif
	}
}

#endif
