#include "config.h"
#include "fep.h"
#include "functions.h"
#include <ctype.h>
#include "ctrlcode.h"

#ifdef USE_SERVER
/* #define SHOW_SERVER_RESPONSE	/* Display the query to server */
#endif

#define WORD_BUF_LEN 128
#define OKURI_LEN 8
#ifdef NUMCONV
# define NUM_BUF_SIZE 10	/* max number of numeric strings */
#endif /* NUMCONV */

#define VOWEL(c) ((c)=='a'||(c)=='i'||(c)=='u'||(c)=='e'||(c)=='o')

static char WordBuf[WORD_BUF_LEN];
static int WordBufLen;
static CandList CurrentCand;
static CandList *FirstCandEntry;
static short OkuriInput;
static char OkuriBuf[OKURI_LEN];
static int OkuriBufLen;

#ifdef NUMCONV
/* for numeric conversion */
# define numConvWrites(word)		numConvWriteOrg((word), writes)
# define numConvWriteShells(word)	numConvWriteOrg((word), writeShells)
static char *numSearchKey();
static int numConv();
static void numConvWriteOrg();
static int numConvStrlen();
#else /* NUMCONV */
# define numSearchKey(word)	(word)
# define numConvStrlen		strlen
# define numConvWrites		writes
# define numConvWriteShells	writeShells
#endif /* NUMCONV */

#ifdef REGMODE
/* for register mode */
static CandList regModeEnter();
int Registering = 0;
#endif /* REGMODE */

#ifdef USE_SERVER
static CandList NetCand;
static int CandFromServer;
LearnMode NetLearnMode = NET_LEARN_MODE;
#endif
#ifdef BOTH_SERVER_AND_SKKSRCH
# define getCandFromServer	getCandFromServerPtr
#endif

#ifdef USERDIC_CHANGED_CHECK
int UserDic_Changed = 0;	/* if not changed, no need to write UserDic */
#endif

extern int writes();
extern int writeShells();
extern int flushOut();
static void putOkuri();
void showCand();

static void
bufferedInput(s)
char *s;
{
	writes(s);
	for (; *s != '\0'; s++)
		WordBuf[WordBufLen++] = *s;
}

void
kkBeg()
{
        toHira();
	setKeymap(&CurrentKeymap,KanjiInputKeymap);
	showmode(KINPUT_MODE);
#ifdef USE_TRIANGLE
	kanjiInputMark(1);
#else
	kanjiInputEffect(1);
#endif
	WordBufLen = 0;
}
void
kkBegV(c)
char c;
{
	flushLastConso('\0',writeShells,flushOut);
	Kindex = 0;
	kkBeg();
	inputKanaVowel(tolower(c),bufferedInput);
}
	
void
kkBegC(c)
char c;
{
	flushLastConso('\0',writeShells,flushOut);
	Kindex = 0;
	kkBeg();
	inputKanaConso(tolower(c),bufferedInput,NULL);
}

void
kkBegA(c)
char c;
{
	flushKana();
	setKeymap(&CurrentKeymap,KAlphaInputKeymap);
	showmode(KALPHA_INPUT_MODE);
#ifdef USE_TRIANGLE
	kanjiInputMark(1);
#else
	kanjiInputEffect(1);
#endif
	WordBufLen = 0;
}

void
kalpha(c)
char c;
{
	write1(c);
	WordBuf[WordBufLen++] = c;
}

void
kaBS(c)
char c;
{
	if (WordBufLen > 0) {
#ifndef USE_TRIANGLE
		kanjiInputEffect(0);
#endif
		rubout(1);
		WordBufLen--;
#ifndef USE_TRIANGLE
		kanjiInputEffect(1);
#endif
	}
	else	/* WordBufLen <= 0 */
		kfCancel();
}

void
kKanaV(c)
char c;
{
	inputKanaVowel(c,bufferedInput);
}

void
okKanaV(c,first)
char c,first;
{
	if (first) {
		WordBuf[WordBufLen++] = c;
	}
	OkuriFirst = 0;
	/* rubout(1); */	/* for '*' */
	inputKanaVowel(tolower(c),putOkuri);
	kkconv();
}

void
kKanaC(c)
char c;
{
	inputKanaConso(c,bufferedInput,NULL);
}

void
okKanaC(c,first)
char c,first;
{
	c = tolower(c);
	OkuriFirst = 0;
	if (Nconso == 0 && OkuriBufLen == 0) {
		WordBuf[WordBufLen++] = c;
	}
	switch (inputKanaConso(c,putOkuri,NULL)) {
	case 0:	/* if double conso, use only newer one (except 'n') */
		if (OkuriBufLen == 0)
			WordBuf[WordBufLen-1] = c; /* replace previous conso */
		break;	/* if typed 'YoMb', Buf should be "Yo*b"(not "Yo*m")*/
	case 1:	/* notOverwrite == 1: for example: ch, ?y */
	case 2: /* for 'n' not replace */
	default:
		break;
	}
}

void
kZenAl(c)
char c;
{
	/* same as flushKana() in iZenAl() but not flushOut by writeShells() */
	flushLastConso('\0',bufferedInput,NULL);
	Kindex = 0;
	SmallTU = 0;
	inputZenkakuAlpha(c,bufferedInput);
}

void
kZenEx(c)
char c;
{
	/* same as flushKana() in iZenEx() but not flushOut by writeShells() */
	flushLastConso('\0',bufferedInput,NULL);
	Kindex = 0;
	SmallTU = 0;
	inputZenkakuEx(c,bufferedInput);
}

void
kfthru(c)
char c;
{
	static char buf[2];
	flushLastConso(c,bufferedInput,NULL);	/* flush 'n' */
	buf[0] = c;
	buf[1] = '\0';
	bufferedInput(buf);
}

void
fxthru(c)
char c;
{
	/* fix and through */
	fixIt();
	(*CurrentKeymap[c])(c);
}

void
kfCancel(c)
char c;
{
#ifndef USE_TRIANGLE
	kanjiInputEffect(0);
#endif
	cancelConso();
	rubout(WordBufLen);
	WordBuf[WordBufLen] = '\0';
#ifdef USE_TRIANGLE
	kanjiInputMark(0);
#endif
	endKanjiInput();
}

void
kfFix(c)
char c;
{
	cancelConso();	/* to clear 'm' when typed 'M^J' */
#ifdef USE_TRIANGLE
	rubout(WordBufLen);
#else
	csrLeft(WordBufLen);
#endif
	WordBuf[WordBufLen] = '\0';
#ifdef USE_TRIANGLE
	kanjiInputMark(0);
#else
	kanjiInputEffect(0);
#endif
	writeShells(WordBuf);
	flushOut(WordBufLen);
	endKanjiInput();
}

void
kfFixToAsc(c)
char c;
{
	kfFix(c);
	toAsc(c);
}

void
kfFixToZenA(c)
char c;
{
	kfFix(c);
	toZenA(c);
}

void
kfFixThru(c)
char c;
{
	kfFix(c);
	thru(c);
}

void
thruKfFixToAsc(c)
char c;
{
	kfFix(c);
	toAsc(c);
	thru(c);
}

void
thruKfFixToEsc(c)
char c;
{
	kfFix(c);
	toEsc(c);
	thru(c);
}

void
okfFix(c)
char c;
{
	cancelOkuri();
	if (CurrentKeymap == OkuriInputKeymap)
		cancelOkuri();
	kfFix(0);
}

void
okfFixToAsc(c)
char c;
{
	okfFix(c);
	toAsc(c);
}

void
okfFixToZenA(c)
char c;
{
	okfFix(c);
	toZenA(c);
}

void
okfFixThru(c)
char c;
{
	okfFix(c);
	thru(c);
}

void
thruOkfFixToAsc(c)
char c;
{
	okfFix(c);
	toAsc(c);
	thru(c);
}

void
thruOkfFixToEsc(c)
char c;
{
	okfFix(c);
	toEsc(c);
	thru(c);
}

void
kfBS(c)
char c;
{
	int i,n;
	char con[MAX_CONSO];

	if (Nconso) {
#ifndef USE_TRIANGLE
		kanjiInputEffect(0);
#endif
		rubout(Nconso);
		for (i = 1; i < Nconso; i++)
			con[i] = LastConso[i];
		n = Nconso;
		Nconso = 0;
		Kindex = 0;
		SmallTU = 0;
		for (i = 1; i < n; i++)
			kKanaC(con[i]);
#ifndef USE_TRIANGLE
		kanjiInputEffect(1);
#endif
	}
	else if (WordBufLen > 0) {
#ifndef USE_TRIANGLE
		kanjiInputEffect(0);
#endif
		if (WordBuf[WordBufLen-1] & 0x80) { /* is KANJI */
			rubout(2);
			WordBufLen -= 2;
		}
		else {
			rubout(1);
			WordBufLen--;
		}
#ifndef USE_TRIANGLE
		kanjiInputEffect(1);
#endif
	}
	else	/* WordBufLen <= 0 */
		kfCancel();
}

void
okuriBS(c)
char c;
{
	int i,n;
	char con[MAX_CONSO];

#ifndef USE_TRIANGLE
	kanjiInputEffect(0);
#endif
	n = Nconso;
	for (i = 1; i < n; i++)
		con[i] = LastConso[i];
	if (n <= 1 && OkuriBufLen == 0) { /* must check OuriBufLen for "Nn" */
		cancelOkuri();
	}
	else if (n == 0 && OkuriBufLen > 0) {	/* BS "nn" */
		if (OkuriBuf[OkuriBufLen-1] & 0x80) { /* is KANJI */
			rubout(2);
			OkuriBufLen -= 2;
		}
		else {
			rubout(1);
			OkuriBufLen--;
		}
		OkuriBuf[OkuriBufLen] = '\0';
		if (OkuriBufLen == 0) {
			clearOkuri();
			OkuriFirst = 1;
		}
	}
	else {
		cancelConso();
		if (OkuriBufLen == 0)
			WordBufLen--;	/* cancel conso. */
		for (i = 1; i < n; i++)
			okKanaC(con[i],0);
	}
#ifndef USE_TRIANGLE
	kanjiInputEffect(1);
#endif
}


void
kkconv(c)
char c;
{
	int l;
	DicList originalDicItem = NULL;

#ifndef USE_TRIANGLE
	kanjiInputEffect(0);
#endif
	flushLastConso('\0',bufferedInput,NULL);
	cancelConso();
	if (WordBufLen == 0 || (OkuriInput && WordBufLen == 1)) {
#ifdef USE_TRIANGLE
		kanjiInputMark(0);
#endif
		endKanjiInput();
		return;
	}
	showmode(KSELECT_MODE); 
		
	WordBuf[WordBufLen] = '\0';
#ifdef USE_SERVER
	CandFromServer = 0;
	NetCand = NULL;
#endif
	CurrentCand = getCand(UserDic,numSearchKey(WordBuf));
	if (CurrentCand) {
		originalDicItem = CurrentCand->dicitem;
		if (OkuriInput)
			CurrentCand = searchOkuri(CurrentCand,
						  OkuriBuf,
						  &FirstCandEntry);
		else
			FirstCandEntry = &(CurrentCand->dicitem->cand);
	}
	else {
		FirstCandEntry = NULL;
	}
#ifdef SYSTEM_DIC_NAME
	if (CurrentCand == NULL) {
		CurrentCand = getCand(SystemDic,numSearchKey(WordBuf));
		if (CurrentCand && OkuriInput)
			CurrentCand = searchOkuri(CurrentCand,
						  OkuriBuf,
						  &FirstCandEntry);
	}
#endif
#ifdef USE_SERVER
	if (CurrentCand == NULL && ServerOK) {
#ifdef SHOW_SERVER_RESPONSE
		showmessage("Querying first candidate to server...");
#endif
		CurrentCand = getCandFromServer(numSearchKey(WordBuf));
		if (CurrentCand && OkuriInput)
			CurrentCand = searchOkuri(CurrentCand,
						  OkuriBuf,
						  &FirstCandEntry);
		if (CurrentCand != NULL) {
			CandFromServer = 1;
			NetCand = CurrentCand;
		}
	}
#endif

	l = WordBufLen;
	if (OkuriInput)
		l += strlen(OkuriBuf);
	rubout(l);
#ifdef USE_TRIANGLE
	kanjiInputMark(0);
#endif
	setKeymap(&CurrentKeymap,convertKeymap(&SelectionKeymap));

	if (CurrentCand == NULL) {
#ifdef REGMODE
		if (!Registering && regModeEnter() != NULL)
			;/* register mode started */
		else {	/* already in register mode or failed to enter */
#endif
			/* {{ now not support recursive registering }} */
			if (PreserveOnFailure) {
				bell();
				backToKanjiInput();
			}
			else {
				endKanjiInput();
			}
#ifdef REGMODE
		}
#endif
	}
	else {
		if (CurrentCand->dicitem == NULL && originalDicItem) {
			CurrentCand->dicitem = originalDicItem;
			if (!OkuriInput)
				FirstCandEntry = &(CurrentCand->dicitem->cand);
		}
		showCand();
		BlockTty = 1;
	}
}

static void
putOkuri(s)
char *s;
{
	char *p;
	int l = OkuriBufLen;
	writes(s);
	for (p = s; *p; p++) {
		OkuriBuf[OkuriBufLen++] = *p;
		if (OkuriBufLen >= OKURI_LEN) {
			bell();
			OkuriBuf[l] = '\0';
			OkuriBufLen = l;
			rubout(strlen(s));
			return;
		}
	}
	OkuriBuf[OkuriBufLen] = '\0';
}

toOkuri()
{
	OkuriInput = 1;
	*OkuriBuf = '\0';
	OkuriBufLen = 0;
	setKeymap(&CurrentKeymap, OkuriInputKeymap);
}
	
void
kOkuri(c)
char c;
{
	char okuri = tolower(c);

	if (Nconso != 0 && LastConso[Nconso] != 'n'
	&&  okuri != LastConso[Nconso]) {	/* toOkuri when `KesSuru' */
		if (VOWEL(okuri))
			kKanaV(okuri);
		else
			kKanaC(okuri);
		return;
	}
	if (WordBufLen == 0) {
		/* Recover chattering effect */
		if (VOWEL(okuri)) {
			kKanaV(okuri);
			return;
		}
		kKanaC(okuri);
		return;
#if 0
		cancelConso();
#ifdef USE_TRIANGLE
		kanjiInputMark(0);
#endif
		endKanjiInput();
#ifndef USE_TRIANGLE
		kanjiInputEffect(0);
#endif
		return;
#endif /* 0 */
	}
	toOkuri();
	/* must do flushLastConso() before write1('*') (ex: KanJi) */
	flushLastConso(okuri,bufferedInput,NULL);
	write1('*');
	(*CurrentKeymap[okuri])(okuri,1);
}

void
stSuffix(c)
char c;
{
	fixIt();
	kkBeg();
	kfthru(c);
}

void
stPrefixCv(c)
char c;
{
	if (WordBufLen == 0) {
		kfthru(c);
	}
	else {
		kfthru(c);
		kkconv(' ');
	}
}

void
showCand()
{
#ifdef USE_TRIANGLE
	kanjiSelectionMark(1);
#else
	kanjiSelectionEffect(1);
#endif
	numConvWrites(CurrentCand->candword);
	if (OkuriInput) {
		writes(OkuriBuf);
	}
}

static
showNextCand(nextcand)
CandList nextcand;
{
	int l;

	l = numConvStrlen(CurrentCand->candword);
	if (OkuriInput)
		l += strlen(OkuriBuf);
#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	rubout(l);
#ifndef USE_TRIANGLE
	kanjiSelectionEffect(1);
#endif
	CurrentCand = nextcand;
	numConvWrites(CurrentCand->candword);
	if (OkuriInput)
		writes(OkuriBuf);
}

void
nxCand()
{
	int l;
	CandList nextcand,nnextcand;
#ifdef REGMODE
	CandList regcand;
#endif

	nextcand = CurrentCand->nextcand;
	nnextcand = CurrentCand;
	while (nextcand && nextcand->okuri) {
		nnextcand = nextcand;
		nextcand = nnextcand->nextcand;
	}
	if (nextcand) {
		showNextCand(nextcand);
		return;		/* have done nxCand() */
	}
#ifdef USE_SERVER
	else if (ServerOK && !CandFromServer) {
#ifdef SHOW_SERVER_RESPONSE
		showmessage("Querying next candidate to server...");
#endif
		NetCand = getCandFromServer(numSearchKey(WordBuf));
		if (NetCand) {
			NetCand = deleteCand(NetCand,firstCand(CurrentCand));
			if (NetCand) {
				CandFromServer = 1;
				nnextcand->nextcand = NetCand;
				NetCand->prevcand = nnextcand;
				if (OkuriInput) {
					showNextCand(searchOkuri(NetCand,
								 OkuriBuf,
								 &FirstCandEntry));
				}
				else {
					showNextCand(NetCand);
				}
				return;	/* have done nxCand() */
			}
		}
	}
#endif
#ifdef REGMODE
	/* not yet showNextCand(). (There is no Cand to show) */
	if (!Registering) {	/* enter register mode */
		l = numConvStrlen(CurrentCand->candword);
		if (OkuriInput)
			l += strlen(OkuriBuf);
#ifndef USE_TRIANGLE
		kanjiSelectionEffect(0);
#endif
		rubout(l);
#ifdef USE_TRIANGLE
		kanjiSelectionMark(0);
#endif
		/* setup some vars (allocate RegisterCand and return it) */
		if ((regcand = regModeEnter()) == NULL) {  /* init failed */
			/* Because regModeEnter() display original WordBuf,
			 * must clear CurrentCand before call regModeEnter().
			 * But regModeEnter() may fail,
			 * then re-display CurrentCand.
			 * {{ Is it better to call regSetupVars() here
			 *    to check whether be able to malloc memory? }}
			 */
#ifdef USE_TRIANGLE
			kanjiSelectionMark(1);	/* re-display */
#else
			kanjiSelectionEffect(1);
#endif
			numConvWrites(CurrentCand->candword);
			if (OkuriInput)
				writes(OkuriBuf);
			return;
		}

		regcand->dicitem = CurrentCand->dicitem;
		/* append RegisterCand to list for regModeCancel() */
		/*nnextcand->nextcand = regcand;*/
		regcand->prevcand = nnextcand;
	}
#endif /* REGMODE */
}

void
pvCand()
{
	int l;
	CandList myCand;

	myCand = CurrentCand;
	if (CurrentCand->prevcand && CurrentCand->prevcand->okuri) {
		CurrentCand = CurrentCand->prevcand;
		pvCand();
	}
	if (CurrentCand == NULL)
		CurrentCand = myCand;
	l = numConvStrlen(CurrentCand->candword);
	if (OkuriInput)
		l += strlen(OkuriBuf);
#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	rubout(l);
	if (CurrentCand->prevcand) {
#ifndef USE_TRIANGLE
		kanjiSelectionEffect(1);
#endif
		CurrentCand = CurrentCand->prevcand;
		numConvWrites(CurrentCand->candword);
		if (OkuriInput)
			writes(OkuriBuf);
	}
	else {
#ifdef USE_TRIANGLE
		kanjiSelectionMark(0);
#endif
		backToKanjiInput();
	}
}

/* back to kanji input mode */
backToKanjiInput()
{
#ifdef USE_SERVER
	if (CandFromServer) {
		if (NetCand->prevcand)
			NetCand->prevcand->nextcand = NULL;
		freeCand(NetCand);
		NetCand = NULL;
		CurrentCand = NULL;
		CandFromServer = 0;
	}
#endif
	restoreKeymap(&CurrentKeymap);
	showmode(KINPUT_MODE);
#ifdef USE_TRIANGLE
	kanjiInputMark(1);
#else
	kanjiInputEffect(1);
#endif
	if (OkuriInput) {
		clearOkuri();
		writes(WordBuf);
		write1('*');
		OkuriFirst = 1;
	}
	else
		writes(WordBuf);
}

void
cancelOkuri()
{
#ifndef USE_TRIANGLE
	kanjiInputEffect(0);
#endif
	if (Nconso == 0 && OkuriBufLen == 0) {
		rubout(1);	/* rub out '*' */
		OkuriInput = 0;
		setKeymap(&CurrentKeymap, KanjiInputKeymap);
	}
	else {
		cancelConso();
		if (OkuriBufLen)
			rubout(OkuriBufLen);
		clearOkuri();
		OkuriFirst = 1;
	}
#ifndef USE_TRIANGLE
	kanjiInputEffect(1);
#endif
}

clearOkuri()
{
	WordBufLen--;
	WordBuf[WordBufLen] = '\0';
	toOkuri();
}

void
fixIt(c)
char c;
{
	int l;
	DicList dlist;

#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	if (CurrentCand != NULL) {
		l = numConvStrlen(CurrentCand->candword);
		if (OkuriInput)
			l += strlen(OkuriBuf);
#ifdef USE_TRIANGLE
		rubout(l);
		kanjiSelectionMark(0);
#else
		csrLeft(l);
#endif
		numConvWriteShells(CurrentCand->candword);
		if (OkuriInput) {
			writeShells(OkuriBuf);
		}
		flushOut(l);
#ifdef USE_SERVER
		if (!CandFromServer) {
			selectCand(FirstCandEntry,CurrentCand);
		}
		else if (NetLearnMode != LearnOff) {
			if (CurrentCand->dicitem == NULL) {
				/*
				 * Candidate from server and no synonym entry
				 * in the personal dictionary
				 */
				if (FirstCandEntry == NULL) {
					dlist = addNewItem(UserDic,numSearchKey(WordBuf),NetCand);
					if (dlist == NULL) { /*malloc failed?*/
						NetLearnMode = LearnOff;
						goto end;
					}
					CurrentCand->dicitem = dlist;
					FirstCandEntry = &(dlist->cand);
				}
				else
					CurrentCand->dicitem = (*FirstCandEntry)->dicitem;
			}
			if (NetCand == CurrentCand) {
				/*
				 * NetCand becomes NetCand->nextcand
				 * not to free the selected candidate.
				 */
				NetCand = CurrentCand->nextcand;
				if (NetCand == NULL) {
					CandFromServer = NIL;
				}
			}
#ifdef USERDIC_CHANGED_CHECK
			UserDic_Changed |= UDIC_SERVER;
#endif
			selectCand(FirstCandEntry,CurrentCand);
		}
#endif
	}
#ifdef USE_TRIANGLE
	else {
		kanjiSelectionMark(0);
	}
#endif
    end:
	endKanjiInput();
}

void
thruFixItToAsc(c)
char c;
{
	fixIt(c);
	toAsc(c);
	thru(c);
}

void
thruFixItToEsc(c)
char c;
{
	fixIt(c);
	toEsc(c);
	thru(c);
}

void
cancelSel(c)
char c;
{
	int l;
	DicList dlist;

#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	if (CurrentCand != NULL) {
		l = numConvStrlen(CurrentCand->candword);
		if (OkuriInput)
			l += strlen(OkuriBuf);
		rubout(l);
	}
#ifdef USE_TRIANGLE
	kanjiSelectionMark(0);
#endif
	backToKanjiInput();
}
	

endKanjiInput()
{
	*OkuriBuf = '\0';
	OkuriBufLen = 0;
	OkuriInput = 0;
	toKana();
	BlockTty = 0;
#ifdef USE_SERVER
	if (CandFromServer) {
		if (NetCand->prevcand)
			NetCand->prevcand->nextcand = NULL;
		freeCand(NetCand);
		NetCand = NULL;
		CurrentCand = NULL;
		CandFromServer = 0;
	}
#endif
}

void
h2kkana(c)
char c;
{
	int l;

#ifndef USE_TRIANGLE
	kanjiInputEffect(0);
#endif
	flushLastConso('\0',bufferedInput,NULL);
	cancelConso();
	if (WordBufLen == 0 || (OkuriInput && WordBufLen == 1)) {
#ifdef USE_TRIANGLE
		kanjiInputMark(0);
#endif
		endKanjiInput();
		return;
	}
	WordBuf[WordBufLen] = '\0';

	l = WordBufLen;
	if (OkuriInput)
		l += strlen(OkuriBuf)-1;
	rubout(l);
#ifdef USE_TRIANGLE
	kanjiInputMark(0);
#endif
	hira2kata(WordBuf);
	writeShells(WordBuf);
	endKanjiInput();
}

#ifdef NUMCONV
/*========== numeric convertion ==========*/
/* for numeric conversion */
/* to do convert word where number as parameter,
 * must search candidates with word in which number is represented as '#'.
 * (ex: "12:00" -> "#:#")
 * (for this purpose, use NumWordBuf[])
 * and must use this NumWordBuf[] when register to dictionary.
 *
 * when print candidate, must real conversion from #0, #1, ... to
 * number in proper format.
 * (ex: "#0:#0" -> "12:00")
 * to do so, we need buffer to hold input number (NumBufStr[]).
 */
static char NumBufStr[WORD_BUF_LEN];
static char *NumBufp[NUM_BUF_SIZE];	/* pointer to in NumBufStr[] */
static int NumBufSize;
static char NumWordBuf[WORD_BUF_LEN];	/* numReplace()'ed WordBuf */

/* check whether word has number (digit characters)
 * Return: 0 if no number, otherwise !0
 */
static int
numExist(word)
char *word;
{
	for (; *word != '\0'; word++) {
		if (isdigit((unsigned char)*word))
			return 1;
	}
	return 0;
}

/* compute search key word and setup some variables for real conversion.
 * replace number string in word with one '#' (ex: "123m" -> "#m").
 * replaced word is written on NumWordBuf.
 * save number in NumBufStr (save number strings separated by '\0'),
 * and get pointer in NumBufp to number string in NumBufStr.
 * Return: number of number string which exists in word.
 * Set:	in NumWordBuf: word where number string is replaced with '#'
 *	pointer to number string is stored in NumBufp[0]
 *	to NumBufp[NumBufSize-1].
 */
static int
numComputeSearchKey(word)
char *word;
{
	static char cachedword[WORD_BUF_LEN];
	static char cachedreplword[WORD_BUF_LEN];
	int i;
	char *replword, *bufp;

	/* initialize: clear cache */
	if (word == NULL) {
		cachedword[0] = '\0';
		cachedreplword[0] = '\0';
		NumBufSize = 0;
		return -1;
	}
	/* can use word computed before? */
	if (strcmp(word, cachedword) == 0) {	/* no need to compute now */
		strcpy(NumWordBuf, cachedreplword);	/* use cached string */
		return NumBufSize;
	}

	/* do computation */
	replword = NumWordBuf;
	bufp = NumBufStr;
	for (i = 0; i < NUM_BUF_SIZE; i++) {
		while (!isdigit((unsigned char)*word) && *word != '\0')
			*replword++ = *word++;	/* skip non digit */
		if (*word == '\0')
			break;
		NumBufp[i] = bufp;	/* start point of number string */
		while (isdigit((unsigned char)*word))
			*bufp++ = *word++;/* save number string on numbufstr */
		*bufp++ = '\0';
		*replword++ = '#';	/* put '#' instead of number string */
	}
	*replword = '\0';
	numConv(NULL, NULL);	/* clear internal cache for real conversion */

	/* setup cache data for future call */
	strcpy(cachedreplword, replword);
	strcpy(cachedword, word);

	NumBufSize = i;
	return NumBufSize;
}

/* return search key word considering numeric conversion.
 * if numeric digits exist in word, do computation search key word and
 * return computed search key (NumWordBuf).
 * (set some variables for real numeric conversion).
 * else only return original word.
 */
static char *
numSearchKey(word)
char *word;
{
	if (numExist(word)) {
		numComputeSearchKey(word);
		return NumWordBuf;
	} else {
		/* It should be better to call numComputeSearchKey()
		 * instead of setting NumBufSize = 0 direct,
		 * because numComputeSearchKey() setups some variables and
		 * has internal cache.
		 */
		numComputeSearchKey(NULL);  /* clear internal cache and vars */
		return word;
	}
}

extern char *ZenkakuAlpha[], *NumKanji[], *NumKurai[], *NumKurai4[];

/* convert ASCII number character to Zenkaku number character */
static char *
numZenkaku(buf, numstr)
char *buf, *numstr;
{
	for (; *numstr != '\0'; numstr++) {
		strcat(buf, ZenkakuAlpha[*numstr - ' ']);
	}
	return buf;
}

/* convert ASCII number character to Kanji number character */
static char *
numKanji(buf, numstr)
char *buf, *numstr;
{
	for (; *numstr != '\0'; numstr++) {
		strcat(buf, NumKanji[*numstr - '0']);
	}
	return buf;
}

/* convert ASCII number character to Kanji number character with KURAI DORI */
static char *
numKanjiKurai(buf, numstr)
char *buf, *numstr;
{
	int len, wrote;

	wrote = 0;
	for (len = strlen(numstr); *numstr != '\0'; numstr++, len--) {
		switch (len % 4) {
		case 1:	/* ICHI */
			if (*numstr != '0') {
				strcat(buf, NumKanji[*numstr - '0']);
				wrote++;
			}
			/* some character is written on this kurai4 */
			if (wrote)	/* MAN, OKU, CHO, KEI, GAI, ... */
				strcat(buf, NumKurai4[len / 4]);
			wrote = 0;
			break;
		case 2:	/* JU */
		case 3:	/* HYAKU */
		case 0:	/* SEN */
			if (*numstr == '0')
				continue;
			if (*numstr != '1')
				strcat(buf, NumKanji[*numstr - '0']);
			strcat(buf, NumKurai[len % 4]);
			wrote++;
			break;
		default:
			break;
		}
	}
	return buf;
}

/* do number conversion
 * #0, #1, #2, #3, #4 to number
 * converted string is written on convword.
 * (convword must have enough space to write on)
 * Return: number of characters written on convword.
 */
static int
numConv(word, convword)
char *word, *convword;
{
	static char origword[WORD_BUF_LEN];	/* orig word of prevword */
	static char prevconv[BUFSIZ];	/* word converted before (cache) */
	int i, convtype;
	char *p, *q;
	char *bufp;

	/* initialize: clear cache */
	if (word == NULL) {
		origword[0] = '\0';
		prevconv[0] = '\0';
		return -1;
	}
	/* can use word converted before? */
	if (strcmp(word, origword) == 0) {	/* no need to convert now */
		strcpy(convword, prevconv);	/* use cached string */
		return strlen(convword);
	}
	/* do conversion */
	p = q = word;
	bufp = convword;
	for (i = 0; i < NumBufSize; i++) {
		while (*p != '#' && *p != '\0')
			*bufp++ = *p++;
		if (*p == '\0')
			break;
		*bufp = '\0';
		/* strtol() after '#' (*p == '#')*/
		q = p + 1;
		convtype = (!isdigit((unsigned char)*q)) ? -1 : strtol(q, &q, 10);
		switch (convtype) {
		case -1:	/* not specification of conversion type */
			*bufp++ = '#';	/* only copy '#' */
			*bufp = '\0';
			break;
		case 0:		/* no conversion */
			strcpy(bufp, NumBufp[i]);
			break;
		case 1:		/* zenkaku number character */
			numZenkaku(bufp, NumBufp[i]);
			break;
		case 2:		/* kanji number character */
			numKanji(bufp, NumBufp[i]);
			break;
		case 3:		/* kanji number character with KURAI DORI */
			numKanjiKurai(bufp, NumBufp[i]);
			break;
		/* {{ not support conversion type 4 and 9 }} */
		case 4:		/* search dictionary with this numeric string*/
		case 9:		/* SHOGI */
		default:	/* not support this conversion type */
			strncpy(bufp, p, q-p);	/* no conversion */
			*(bufp + (q-p)) = '\0';
			break;
		}
		while (*bufp != '\0')
			bufp++;	/* set bufp to end of buffer */
		p = q;
	}
	while (*p != '\0')		/* trailing string */
		*bufp++ = *p++;
	*bufp = '\0';

	/* setup cache data for future call */
	strcpy(prevconv, convword);
	strcpy(origword, word);

	return bufp - convword;
}

/* write word after convert number.
 * as write function, use paremeter output.
 * if no need to convert, only output(word).
 */
static void
numConvWriteOrg(word, output)
char *word;
void (*output)();
{
	char convword[BUFSIZ];

	if (NumBufSize) {
		numConv(word, convword);
		output(convword);
	} else {
		output(word);
	}
}

/* return strlen of word after convert number */
/* if no need to convert, only return strlen(word) */
static int
numConvStrlen(word)
char *word;
{
	char convword[BUFSIZ];

	if (NumBufSize) {
		numConv(word, convword);
		return strlen(convword);
	} else {
		return strlen(word);
	}
}
#endif /* NUMCONV */

#ifdef REGMODE
/*========== register mode ==========*/
static char RegisterWord[WORD_BUF_LEN];
static CandList RegisterCand;
static int RegEditPos = 0;	/* current cursor position */
/* to save old status */
static CandList *RegFirstCandEntry = NULL;
static short RegOkuriInput;
static char RegOkuriBuf[OKURI_LEN];
static char RegOrigWordBuf[WORD_BUF_LEN];  /* original input word for number */
static KeymapPtr RegKinputKeymap = KanjiInputKeymap;
#ifdef USE_SERVER
static CandList RegNetCand;
static int RegCandFromServer;
#endif

/* setup variables to enter register mode */
static int
regSetupVars()
{
	Registering = 1;
	/* (see regRestoreVars() to restore NumWordBuf and WordBuf) */
	strcpy(RegisterWord, numSearchKey(WordBuf));
	strcpy(RegOrigWordBuf, WordBuf);	/* save original WordBuf */
	RegFirstCandEntry = FirstCandEntry;
	RegOkuriInput = OkuriInput;
	strcpy(RegOkuriBuf, OkuriBuf);
#ifdef USE_SERVER
	RegNetCand = NetCand;
	RegCandFromServer = CandFromServer;
	if (CandFromServer) {
		if (NetCand->prevcand)
			NetCand->prevcand->nextcand = NULL;	/* cut link */
	}
#endif
	RegEditPos = 0;
	/* save to cancel (CurrentKeymap is SelectionKeymap) */
	RegKinputKeymap = getLastKeymap();

	RegisterCand = _NEW2(CandList,WORD_BUF_LEN);
	if (RegisterCand == NULL)
		return -1;
	RegisterCand->okuri = NULL;
	RegisterCand->nextcand = NULL;
	RegisterCand->prevcand = NULL;
	RegisterCand->dicitem = NULL;
	RegisterCand->candword[0] = '\0';
}

/* restore saved variables when registe mode ended */
static void
regRestoreVars()
{
	Registering = 0;
	CurrentCand = RegisterCand;
	/* restore buffers */
	strcpy(WordBuf, RegOrigWordBuf);/*if Num, RegisterWord was NumWordBuf*/
	WordBufLen = strlen(WordBuf);
	FirstCandEntry = RegFirstCandEntry;
	OkuriInput = RegOkuriInput;
	strcpy(OkuriBuf, RegOkuriBuf);
	OkuriBufLen = strlen(OkuriBuf);
#ifdef USE_SERVER
	NetCand = RegNetCand;
	CandFromServer = RegCandFromServer;
	if (CandFromServer) {
		if (NetCand->prevcand)		/* re-connect link */
			NetCand->prevcand->nextcand = NetCand;
	}
#endif
	numSearchKey(WordBuf);	/* re-compute search key and setup vars */
}


/* clear variables for register mode */
static void
regClearVars(isfree)
int isfree;
{
	/* clear variables for register mode */
	if (isfree && RegisterCand != NULL) {
		RegisterCand->prevcand = NULL;	/* nextcand should be NULL */
		freeCand(RegisterCand);
	}
	RegisterCand = NULL;
	RegisterWord[0] = '\0';
	RegOrigWordBuf[0] = '\0';
	RegFirstCandEntry = NULL;
	RegOkuriInput = 0;
	RegOkuriBuf[0] = '\0';
#ifdef USE_SERVER
	RegNetCand = NULL;
	RegCandFromServer = 0;
#endif
	RegEditPos = 0;
	RegKinputKeymap = KanjiInputKeymap;
}

/* enter register mode (setup some functions and print registering YOMI) */
static CandList
regModeEnter()
{
	extern char RegisterMark[];

	if (regSetupVars() < 0)
		return NULL;
#ifdef USE_TRIANGLE
	kanjiSelectionMark(1);	/* show begining of RegisterWord */
#else
	kanjiSelectionEffect(1);	/* show begining of RegisterWord */
#endif
	writes(RegisterWord);
	writes(RegisterMark);
	showmode(REGISTER_MODE);

	/* do not call endKanjiInput()
	 * not to free NetCand for regModeCancel().
	 */
	*OkuriBuf = '\0';
	OkuriBufLen = 0;
	OkuriInput = 0;
	toKana();
	BlockTty = 0;
	NetCand = NULL;
	CandFromServer = 0;

	return RegisterCand;
}

/* cancel register mode */
static void
regModeCancel()
{
	extern char RegisterMark[];
	CandList prevcand, pprevcand;
	int len = strlen(RegisterWord) + strlen(RegisterMark)
		+ strlen(RegisterCand->candword);

#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	rubout(len);
#ifdef USE_TRIANGLE
	kanjiSelectionMark(0);
#endif
	regRestoreVars();
	/* set keymap to KanjiInputKeymap or KAlphaInputKeymap */
	/* to make restoreKeymap() work well in backToKanjiInput() */
	setKeymap(&CurrentKeymap, RegKinputKeymap);

	/* need to back to selection mode? */
	prevcand = CurrentCand->prevcand;
	pprevcand = CurrentCand;
	while (prevcand && prevcand->okuri) { /* skip Cand that have okuri */
		pprevcand = prevcand;
		prevcand = pprevcand->prevcand;
	}
	if (prevcand) {		/* back to selection mode */
		setKeymap(&CurrentKeymap, convertKeymap(&SelectionKeymap));
		CurrentCand = prevcand;
		CurrentCand->nextcand = NULL;	/* cancel RegisterCand */
		showCand();
	} else {
		/* dummy setKeymap() to make restoreKeymap() work well */
		setKeymap(&CurrentKeymap, KanjiInputKeymap);
		backToKanjiInput();
	}
	regClearVars(1);
}

/* do register */
static void
registerIt()
{
	extern char RegisterMark[];
	int len;
	
	if ((len = strlen(RegisterCand->candword)) == 0) {
		regModeCancel();
		return;
	}
	len += strlen(RegisterWord) + strlen(RegisterMark);
#ifndef USE_TRIANGLE
	kanjiSelectionEffect(0);
#endif
	rubout(len);
#ifdef USE_TRIANGLE
	kanjiSelectionMark(0);
#endif
	regRestoreVars();
	numConvWriteShells(CurrentCand->candword);
	len = numConvStrlen(CurrentCand->candword);
	if (OkuriInput) {
		writeShells(OkuriBuf);
		len += strlen(OkuriBuf);
	}
	flushOut(len);

	if (CurrentCand->dicitem == NULL) {
		/* no synonym entry in the personal dictionary */
		if (FirstCandEntry == NULL) {
			/* Maybe this word is fixIt()'ed and addNewItem()'ed
			 * by getting cand from Server */
			CandList tmpCand = getCand(UserDic,RegisterWord);
			if (tmpCand != NULL) {
				CurrentCand->dicitem = tmpCand->dicitem;
				FirstCandEntry = &(tmpCand->dicitem->cand);
			} else {
				DicList dlist;
				dlist = addNewItem(UserDic,RegisterWord,CurrentCand);
				if (dlist == NULL)	/* malloc failed? */
					goto end;
				CurrentCand->dicitem = dlist;
				FirstCandEntry = &(dlist->cand);
			}
		}
		else
			CurrentCand->dicitem = (*FirstCandEntry)->dicitem;
	}		
#ifdef USERDIC_CHANGED_CHECK
	UserDic_Changed |= UDIC_REGISTER;
#endif
	selectCand(FirstCandEntry,CurrentCand);

    end:
	regClearVars(0);
	endKanjiInput();
}

/* go to top of line */
static int
regEditTop(word, pos)
char *word;
int pos;
{
	csrLeft(pos);
	return 0;
}

/* go to end of line */
static int
regEditEnd(word, pos)
char *word;
int pos;
{
	int len = strlen(word);
	csrRight(len - pos);
	return len;
}

/* go back */
static int
regEditBack(word, pos)
char *word;
int pos;
{
	int width = (word[pos - 1] & 0x80)? 2: 1;

	if (pos <= 0 || pos - width < 0)
		return pos;
	csrLeft(width);
	return pos - width;
}

/* go forward */
static int
regEditForward(word, pos)
char *word;
int pos;
{
	int width = (word[pos] & 0x80)? 2: 1;
	int len = strlen(word);

	if (pos + width > len)
		return pos;
	csrRight(width);
	return pos + width;
}

/* backspace */
/* Return: new position in word */
static int
regEditBS(word, pos)
char *word;
int pos;
{
	char *p, *curp = word + pos;
	int dellen, newpos, movelen = strlen(curp);

	if ((newpos = regEditBack(word, pos)) == pos)
		return pos;	/* cannot back */
	dellen = pos - newpos;
#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_delchars(dellen);
#else
	if (*curp != '\0')
		writes(curp);	/* now cursor is at end of word */
	erase(dellen);		/* erase trailing string */
	csrLeft(dellen + movelen);	/* reposition cursor */
#endif

#if HAVE_MEMMOVE
	memmove(curp - dellen, curp, movelen + 1);	/* +1 for '\0' */
#else
	for (p = curp - dellen; *curp != '\0'; )
		*p++ = *curp++;
	*p = *curp;	/* '\0' */
#endif
	return newpos;
}

/* delete char */
/* Return: new position in word */
static int
regEditDel(word, pos)
char *word;
int pos;
{
	int movelen, dellen;
	char *p, *newp;

	if (word[pos] == '\0')
		return pos;
	dellen = (word[pos] & 0x80)? 2: 1;
	newp = word + pos + dellen;
	movelen = strlen(newp);
#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_delchars(dellen);
#else
	if (*newp != '\0')
		writes(newp);	/* now cursor is at end of word */
	erase(dellen);		/* erase trailing string */
	csrLeft(dellen + movelen);	/* reposition cursor */
#endif

#if HAVE_MEMMOVE
	memmove(newp - dellen, newp, movelen + 1);	/* +1 for '\0' */
#else
	for (p = newp - dellen; *newp != '\0'; )
		*p++ = *newp++;
	*p = *newp;	/* '\0' */
#endif
	return pos;
}

/* kill to end of line */
static int
regEditKill(word, pos)
char *word;
int pos;
{
	char *curp = word + pos;
	int len = strlen(curp);

#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_delchars(len);
#else
	erase(len);
	csrLeft(len);
#endif
	word[pos] = '\0';
	return pos;
}

/* clear to start of line */
static int
regEditClear(word, pos)
char *word;
int pos;
{
	char *curp = word + pos;
	int len = strlen(word);
	int movelen = strlen(curp);
	int newpos;

	if ((newpos = regEditTop(word, pos)) == pos)
		return pos;	/* nothing to clear */
#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_delchars(len - movelen);
#else
	if (*curp != '\0')
		writes(curp);
	erase(len - movelen);	/* erase trailing string */
	csrLeft(len);	/* go to point of curp (top of line) */
#endif
#if HAVE_MEMMOVE
	memmove(word, curp, movelen + 1);	/* +1 for '\0' */
#else
	while (*curp != '\0')
		*word++ = *curp++;
	*word = *curp;	/* '\0' */
#endif
	return newpos;
}

/* insert string */
static int
regEditInsert(word, pos, str)
char *word, *str;
int pos;
{
	char *curp = word + pos;
	char *src, *dst;
	int movelen = strlen(curp);
	int slen = strlen(str);

#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_inschars(str);
#else
	if (*str != '\0')
		writes(str);
	if (*curp != '\0')
		writes(curp);
	csrLeft(movelen);
#endif

#if HAVE_MEMMOVE
	memmove(curp + slen, curp, movelen + 1);	/* +1 for '\0' */
	memcpy(curp, str, slen);
#else
	/* may overlap */
	for (src = curp + movelen, dst = src + slen; src != curp;)
		*dst-- = *src--;
	*dst = *src;	/* copy *curp */
	for (dst = curp; *str != '\0'; )
		*dst++ = *str++;
	/* must not copy '\0' of str */
#endif
	return pos + slen;
}

/* update inputted string. if CTRL_M, do register */
void
regUpdateCand1(c)
char c;
{
	int len = strlen(RegisterCand->candword);

	switch (c) {
	case CTRL_M:	/* do register */
		if (RegEditPos < len)	/* to clear cand */
			regEditEnd(RegisterCand->candword, RegEditPos);
		registerIt();
		break;
	case CTRL_G:	/* cancel register mode */
	/* case ESC_CODE:	ignore ESC */
		if (RegEditPos < len)	/* to clear cand */
			regEditEnd(RegisterCand->candword, RegEditPos);
		regModeCancel();
		break;
	case CTRL_A:	/* top of line */
		RegEditPos = regEditTop(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_E:	/* end of line */
		RegEditPos = regEditEnd(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_B:	/* back RegEditPos */
		if (RegEditPos > 0)
			RegEditPos = regEditBack(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_F:	/* forward RegEditPos */
		if (RegEditPos < len)
			RegEditPos = regEditForward(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_H:	/* BackSpace */
		if (RegEditPos > 0)
			RegEditPos = regEditBS(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_D:	/* delete char */
	case DEL_CODE:	/* delete char */
		if (RegEditPos < len - 1)
			RegEditPos = regEditDel(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_K:	/* kill to end of line */
		if (RegEditPos < len)
			RegEditPos = regEditKill(RegisterCand->candword, RegEditPos);
		break;
	case CTRL_U:	/* clear to start of line */
		if (RegEditPos > 0)
			RegEditPos = regEditClear(RegisterCand->candword, RegEditPos);
		break;
	default:
		if (c >= ' ') {	/* insert */
			char tmpbuf[2];
			tmpbuf[0] = c;
			tmpbuf[1] = '\0';
			RegEditPos = regEditInsert(RegisterCand->candword, RegEditPos, tmpbuf);
		}
		break;
	}
#if defined(SKKFEP_LIB) && !defined(MOVE_CURSOR)
	skkfep_update_screen();
#endif
}

/* update register buffer */
void
regUpdateCands(s)
char *s;
{
	RegEditPos = regEditInsert(RegisterCand->candword, RegEditPos, s);
}
#endif /* REGMODE */
